import typing

from ..rule_manager import RuleParameters


class RuleParameter:
    def __init__(self, rule_parameters: RuleParameters, filename: str, *keys, default_value: typing.Any = 'ToDo'):

        assert isinstance(rule_parameters, RuleParameters), type(rule_parameters)
        self.__parameters: RuleParameters = rule_parameters

        assert isinstance(filename, str)
        self.filename: str = filename
        self.keys = keys

        # as a convenience try to create the file and the file structure
        self.__parameters.add_param(self.filename, *self.keys, default_value=default_value)

    def get_value(self):
        return self.__parameters.get_param(self.filename, *self.keys)

    def __eq__(self, other):
        return self.get_value() == other

    def __lt__(self, other):
        if not isinstance(other, (int, float)):
            return NotImplemented

        return self.get_value() < other

    def __le__(self, other):
        if not isinstance(other, (int, float)):
            return NotImplemented

        return self.get_value() <= other

    def __ge__(self, other):
        if not isinstance(other, (int, float)):
            return NotImplemented

        return self.get_value() >= other

    def __gt__(self, other):
        if not isinstance(other, (int, float)):
            return NotImplemented

        return self.get_value() > other

    def __repr__(self):
        return f'<RuleParameter file: {self.filename}, keys: {self.keys}, value: {self.get_value()}'



import HABApp.util

import HABApp.config
import HABApp.core

import HABApp.mqtt
import HABApp.openhab
import HABApp.rule
import HABApp.runtime

from HABApp.rule import Rule
#from HABApp.runtime import Runtime



from HABApp.core.items import Item
from .. import get_openhab_interface


class DimmerItem(Item):

    def set_state(self, new_state) -> bool:
        if new_state == 'ON':
            new_state = 100
        if new_state == 'OFF':
            new_state = 0

        assert isinstance(new_state, (int, float)) or new_state is None, new_state
        return super().set_state(new_state)

    def on(self):
        """Switch on"""
        get_openhab_interface().send_command(self.name, 'ON')

    def off(self):
        """Switch off"""
        get_openhab_interface().send_command(self.name, 'OFF')

    def percent(self, value: float):
        """Command dimmer to value (in percent)"""
        assert 0 <= value <= 100
        get_openhab_interface().send_command(self.name, str(value))

    def __str__(self):
        return self.state



import HABApp
from . import WrappedFunction
from .events import AllEvents


class EventBusListener:
    def __init__(self, name, callback, event_type=AllEvents):
        assert isinstance(name, str) or name is None, type(name)
        assert isinstance(callback, WrappedFunction)

        self.name: str = name
        self.func = callback

        self.event_filter = event_type

    def notify_listeners(self, event):

        if self.event_filter is AllEvents or isinstance(event, self.event_filter):
            self.func.run(event)
            return None

        # Make it possible to specify multiple classes
        if isinstance(self.event_filter, list) or isinstance(self.event_filter, set):
            for cls in self.event_filter:
                if isinstance(event, cls):
                    self.func.run(event)
                    return None

    def cancel(self):
        """Stop listening on the event bus"""
        HABApp.core.EventBus.remove_listener(self)



from HABApp.core.items import Item
from .. import get_openhab_interface


class SwitchItem(Item):
    ON = 'ON'
    OFF = 'OFF'

    def set_state(self, new_state) -> bool:
        if new_state is not None and new_state != SwitchItem.ON and new_state != SwitchItem.OFF:
            raise ValueError(f'Invalid value for SwitchItem {self.name}: {new_state}')
        return super().set_state(new_state)

    def is_on(self) -> bool:
        """Test value against on-value"""
        return True if self.state == SwitchItem.ON else False

    def is_off(self) -> bool:
        """Test value against off-value"""
        return True if self.state == SwitchItem.OFF else False

    def on(self):
        """Command item on"""
        get_openhab_interface().send_command(self.name, SwitchItem.ON)

    def off(self):
        """Command item off"""
        get_openhab_interface().send_command(self.name, SwitchItem.OFF)

    def __str__(self):
        return self.state

    def __eq__(self, other):
        if isinstance(other, SwitchItem):
            return self.state == other.state
        elif isinstance(other, str):
            return self.state == other
        elif isinstance(other, int):
            if other and self.is_on():
                return True
            if not other and self.is_off():
                return True
            return False

        return NotImplemented

    def __bool__(self):
        return self.is_on()



from .configentry import ConfigEntry, ConfigEntryContainer


class Ping(ConfigEntry):
    def __init__(self):
        super().__init__()
        self.enabled = False
        self.item = ''
        self.interval = 10


class General(ConfigEntry):
    def __init__(self):
        super().__init__()
        self.listen_only = False


class Connection(ConfigEntry):
    def __init__(self):
        super().__init__()
        self.host = 'localhost'
        self.port = 8080
        self.user = ''
        self.password = ''

        self._entry_kwargs['user'] = {'default': ''}
        self._entry_kwargs['password'] = {'default': ''}


class Openhab(ConfigEntryContainer):
    def __init__(self):
        self.ping = Ping()
        self.connection = Connection()
        self.general = General()



import typing
from threading import Lock


class ValueChanger:
    def __init__(self, parent, initial_value=None):

        self.__parent: PrioritizedValue = parent

        self._value = initial_value
        self._enabled = True if initial_value is not None else False

    def set_value(self, value):
        """Set new value and recalculate overall value

        :param value: new value
        """
        self._enabled = True if value is not None else False
        self._value = value

        self.__parent._value_changed(self)

    def set_enabled(self, value: bool):
        """Enable or disable and recalculate overall value

        :param value: True/False
        """
        assert value is True or value is False
        self._enabled = value

        self.__parent._value_changed(self)

    def __repr__(self):
        return f'<{self.__class__.__name__} enabled: {self._enabled}, value: {self._value}>'


class PrioritizedValue:
    """Thread safe value prioritizer"""

    def __init__(self, on_change):
        self.on_value_change = on_change

        self.__value = None

        self.__childs: typing.Dict[int, ValueChanger] = {}
        self.__lock = Lock()

    def get_value_changer(self, priority: int, initial_value=None) -> ValueChanger:
        """ Create a new instance which can be used to set values

        :param priority: priority of the value
        :param initial_value: initial value
        """
        if priority in self.__childs:
            return self.__childs[priority]

        self.__childs[priority] = ret = ValueChanger(self, initial_value)
        return ret

    def _value_changed(self, child):

        # recalculate value
        new_value = None
        with self.__lock:
            for prio, child in sorted(self.__childs.items()):
                assert isinstance(child, ValueChanger)

                if not child._enabled:
                    continue
                new_value = child._value

            value_changed = new_value != self.__value
            self.__value = new_value

        # Notify that the value has changed
        if value_changed:
            self.on_value_change(new_value)



import asyncio


class FinishedProcessInfo:
    """Information about the finished process."""


    def __init__(self, returncode: int, stdout: str, stderr: str):
        self.returncode: int = returncode
        self.stdout: str = stdout
        self.stderr: str = stderr

    def __repr__(self):
        return f'<ProcessInfo: returncode:{self.returncode} stdout:{self.stdout} stderr:{self.stderr}>'


async def async_subprocess_exec(callback, program: str, *args, capture_output=True):
    assert isinstance(program, str), type(program)

    stdout = None
    stderr = None
    ret_code = None

    try:
        proc = await asyncio.create_subprocess_exec(
            program,
            *args,
            stdout=asyncio.subprocess.PIPE if capture_output else None,
            stderr=asyncio.subprocess.PIPE if capture_output else None
        )

        stdout, stderr = await proc.communicate()
        stdout = stdout.decode()
        stderr = stderr.decode()
        ret_code = proc.returncode
    except asyncio.CancelledError:
        if proc is not None:
            proc.terminate()

        stdout = None
        stderr = None
        ret_code = -2
    except Exception as e:
        ret_code = -1
        stderr = str(e)

    callback(FinishedProcessInfo(ret_code, stdout, stderr))
    return None



from .scheduler import ReoccurringScheduledCallback, ScheduledCallback, WorkdayScheduledCallback, WeekendScheduledCallback, DayOfWeekScheduledCallback
from .rule import Rule

from HABApp.rule.interfaces import FinishedProcessInfo


import asyncio

import HABApp.config
import HABApp.core
import HABApp.rule_manager
import HABApp.util

from .folder_watcher import FolderWatcher
from .shutdown_helper import ShutdownHelper


class Runtime:

    def __init__(self):

        self.loop = asyncio.get_event_loop()

        self.shutdown = ShutdownHelper()

        self.folder_watcher: FolderWatcher = FolderWatcher()

        self.config: HABApp.config.Config = None

        self.async_http: HABApp.rule.interfaces.AsyncHttpConnection = HABApp.rule.interfaces.AsyncHttpConnection()

        # OpenHAB
        self.openhab_connection: HABApp.openhab.OpenhabConnection = None

        # MQTT
        self.mqtt_connection: HABApp.mqtt.MqttConnection = None

        # Rule engine
        self.rule_manager: HABApp.rule_manager.RuleManager = None
        self.rule_params: HABApp.rule_manager.RuleParameters = None

        # Async Workers & shutdown callback
        HABApp.core.WrappedFunction._EVENT_LOOP = self.loop
        self.shutdown.register_func(HABApp.core.WrappedFunction._WORKERS.shutdown)

    def load_config(self, config_folder):

        # Start Folder watcher!
        self.folder_watcher.start(self.shutdown)

        self.config = HABApp.config.Config(self, config_folder=config_folder)

        # OpenHAB
        self.openhab_connection = HABApp.openhab.OpenhabConnection(self.config, self.shutdown)

        # MQTT
        self.mqtt_connection = HABApp.mqtt.MqttConnection(self.config.mqtt, self.shutdown)
        self.mqtt_connection.connect()

        # Rule engine
        self.rule_manager = HABApp.rule_manager.RuleManager(self)
        self.rule_params = HABApp.rule_manager.RuleParameters(self.config, self.folder_watcher)

    @HABApp.util.PrintException
    def get_async(self):
        return asyncio.gather(
            self.async_http.create_client(),
            self.openhab_connection.start(),
            self.rule_manager.get_async(),
        )



import colorsys
import typing

from . import Item

HUE_FACTOR = 360
PERCENT_FACTOR = 100


class ColorItem(Item):
    def __init__(self, name: str, h=0.0, s=0.0, v=0.0):
        super().__init__(name=name, state=(h, s, v))

        self.hue: float = min(max(0.0, h), HUE_FACTOR)
        self.saturation: float = min(max(0.0, s), PERCENT_FACTOR)
        self.value: float = min(max(0.0, v), PERCENT_FACTOR)

    def set_state(self, hue=0.0, saturation=0.0, value=0.0):

        # map tuples to variables
        # when processing events instead of three values we get the tuple
        if isinstance(hue, tuple):
            value = hue[2]
            saturation = hue[1]
            hue = hue[0]

        self.hue = min(max(0.0, hue), HUE_FACTOR)
        self.saturation = min(max(0.0, saturation), PERCENT_FACTOR)
        self.value = min(max(0.0, value), PERCENT_FACTOR)

        return super().set_state(new_state=(hue, saturation, value))

    def get_rgb(self, max_rgb_value=255) -> typing.Tuple[int, int, int]:
        r, g, b = colorsys.hsv_to_rgb(
            self.hue / HUE_FACTOR,
            self.saturation / PERCENT_FACTOR,
            self.value / PERCENT_FACTOR
        )
        return int(r * max_rgb_value), int(g * max_rgb_value), int(b * max_rgb_value)

    def set_rgb(self, r, g, b, max_rgb_value=255):
        h, s, v = colorsys.rgb_to_hsv(r / max_rgb_value, g / max_rgb_value, b / max_rgb_value)
        self.hue = h * HUE_FACTOR
        self.saturation = s * PERCENT_FACTOR
        self.value = v * PERCENT_FACTOR
        return self

    def __repr__(self):
        return f'<Color Hue: {self.hue}°, Saturation: {self.saturation}%, Value: {self.value}%>'



import asyncio
import dataclasses
import datetime
import logging
import typing

import HABApp
import HABApp.core
import HABApp.openhab.events
from HABApp.util import PrintException
from . import definitions
from .http_connection import HttpConnection

log = logging.getLogger('HABApp.openhab.Connection')


@dataclasses.dataclass()
class OpenhabItemDefinition:
    type: str
    name: str
    state: typing.Any
    editable: bool
    label = ''
    category = ''
    tags: typing.List[str] = dataclasses.field(default_factory=list)
    groups: typing.List[str] = dataclasses.field(default_factory=list)
    members: 'typing.List[OpenhabItemDefinition]' = dataclasses.field(default_factory=list)

    @classmethod
    def from_dict(cls, data) -> 'OpenhabItemDefinition':
        assert isinstance(data, dict), type(dict)
        data['groups'] = data.pop('groupNames', [])

        # remove link
        data.pop('link', None)

        # map states, quick n dirty
        state = data['state']
        if state == 'NULL':
            state = None
        else:
            try:
                state = int(state)
            except ValueError:
                try:
                    state = float(state)
                except ValueError:
                    pass
        data['state'] = state

        for i, item in enumerate(data.get('members', [])):
            data['members'][i] = cls.from_dict(item)

        # Important, sometimes OpenHAB returns more than in the schema spec, so we remove those items otherwise we
        # get e.g.: TypeError: __init__() got an unexpected keyword argument 'stateDescription'
        return cls(**{k: v for k, v in data.items() if k in cls.__annotations__})


class OpenhabInterface:
    def __init__(self, connection):

        self.__loop = asyncio.get_event_loop()
        self.__connection: HttpConnection = connection

        # build str like this: '+1000' with the current timezone
        timezone_delta = datetime.datetime.now().astimezone().utcoffset()
        hours = int(timezone_delta.total_seconds()) // 3600
        minutes = (int(timezone_delta.total_seconds()) - hours * 3600) // 60
        self._timezone = f'{hours:+03d}{minutes:02d}'

    def __convert_to_oh_type(self, _in):
        if isinstance(_in, datetime.datetime):
            return _in.strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + self._timezone
        elif isinstance(_in, HABApp.core.items.Item):
            return str(_in.state)
        elif isinstance(_in, HABApp.core.items.ColorItem):
            return f'{_in.hue:.1f},{_in.saturation:.1f},{_in.value:.1f}'
        elif isinstance(_in, (set, list, tuple)):
            return ','.join(str(k) for k in _in)

        return str(_in)

    @PrintException
    def post_update(self, item_name: str, state):
        """
        Post an update to the item

        :param item_name: item name or item
        :param state: new item state
        """
        assert isinstance(item_name, (str, HABApp.core.items.Item)), type(item_name)

        if not self.__connection.is_online or self.__connection.is_read_only:
            return None

        if isinstance(item_name, HABApp.core.items.Item):
            item_name = item_name.name

        asyncio.run_coroutine_threadsafe(
            self.__connection.async_post_update(item_name, self.__convert_to_oh_type(state)),
            self.__loop
        )

    @PrintException
    def send_command(self, item_name: str, command):
        """
        Send the specified command to the item

        :param item_name: item name or item
        :param command: command
        """
        assert isinstance(item_name, (str, HABApp.core.items.Item)), type(item_name)

        if not self.__connection.is_online or self.__connection.is_read_only:
            return None

        if isinstance(item_name, HABApp.core.items.Item):
            item_name = item_name.name

        asyncio.run_coroutine_threadsafe(
            self.__connection.async_send_command(item_name, self.__convert_to_oh_type(command)),
            self.__loop
        )

    @PrintException
    def create_item(self, item_type: str, name: str, label="", category="",
                    tags: typing.List[str] = [], groups: typing.List[str] = [],
                    group_type: str = '', group_function: str = '', group_function_params: typing.List[str] = []):
        """Creates a new item in the OpenHAB item registry or updates an existing one

        :param item_type: item type
        :param name: item name
        :param label: item label
        :param category: item category
        :param tags: item tags
        :param groups: in which groups is the item
        :param group_type: what kind of group is it
        :param group_function: Group state aggregation function
        :param group_function_params: params for group state aggregation
        :return: True if item was created/updated
        """
        if not self.__connection.is_online or self.__connection.is_read_only:
            return None

        def validate(_in):
            assert isinstance(_in, str), type(_in)

        # limit values to special entries and validate parameters
        assert item_type in definitions.ITEM_TYPES, \
            f'{item_type} is not an OpenHAB type: {", ".join(definitions.ITEM_TYPES)}'
        assert isinstance(name, str), type(name)
        assert isinstance(label, str), type(label)
        assert isinstance(category, str), type(category)
        map(validate, tags)
        map(validate, groups)
        assert isinstance(group_type, str), type(group_type)
        assert isinstance(group_function, str), type(group_function)
        map(validate, group_function_params)

        if group_type or group_function or group_function_params:
            assert item_type == 'Group', f'Item type must be "Group"! Is: {item_type}'

            if group_function:
                assert group_function in definitions.GROUP_FUNCTIONS, \
                    f'{item_type} is not a group function: {", ".join(definitions.ITEM_TYPES)}'

        fut = asyncio.run_coroutine_threadsafe(
            self.__connection.async_create_item(
                item_type, name,
                label=label, category=category, tags=tags, groups=groups,
                group_type=group_type, group_function=group_function, group_function_params=group_function_params
            ),
            self.__loop
        )
        return fut.result()

    def get_item(self, item_name: str) -> OpenhabItemDefinition:
        """ Return the complete OpenHAB item definition

        :param item_name: name of the item or item
        """
        if isinstance(item_name, HABApp.core.items.Item):
            item_name = item_name.name
        assert isinstance(item_name, str), type(item_name)

        fut = asyncio.run_coroutine_threadsafe(
            self.__connection.async_get_item(item_name),
            self.__loop
        )
        data = fut.result()
        return OpenhabItemDefinition.from_dict(data)

    @PrintException
    def remove_item(self, item_name: str):
        """
        Removes an item from the openHAB item registry

        :param item_name: name
        """
        if not self.__connection.is_online or self.__connection.is_read_only:
            return None

        fut = asyncio.run_coroutine_threadsafe(
            self.__connection.async_remove_item(item_name),
            self.__loop
        )
        return fut.result()

    def item_exists(self, item_name: str):
        """
        Check if an item exists in the OpenHAB item registry

        :param item_name: name
        """
        assert isinstance(item_name, str), type(item_name)
        fut = asyncio.run_coroutine_threadsafe(
            self.__connection.async_item_exists(item_name),
            self.__loop
        )
        return fut.result()

    def set_metadata(self, item_name: str, namespace: str, value: str, config: dict):
        """
        Add/set metadata to an item

        :param item_name: name of the item or item
        :param namespace: namespace
        :param value: value
        :param config: configuration
        :return:
        """
        if isinstance(item_name, HABApp.core.items.Item):
            item_name = item_name.name
        assert isinstance(item_name, str), type(item_name)
        assert isinstance(namespace, str), type(namespace)
        assert isinstance(value, str), type(value)
        assert isinstance(config, dict), type(config)

        fut = asyncio.run_coroutine_threadsafe(
            self.__connection.async_set_metadata(item_name=item_name, namespace=namespace, value=value, config=config),
            self.__loop
        )
        return fut.result()

    def remove_metadata(self, item_name: str, namespace: str):
        """
        Remove metadata from an item

        :param item_name: name of the item or item
        :param namespace: namespace
        :return:
        """
        if isinstance(item_name, HABApp.core.items.Item):
            item_name = item_name.name
        assert isinstance(item_name, str), type(item_name)
        assert isinstance(namespace, str), type(namespace)

        fut = asyncio.run_coroutine_threadsafe(
            self.__connection.async_remove_metadata(item_name=item_name, namespace=namespace),
            self.__loop
        )
        return fut.result()


OH_INTERFACE = None


def get_openhab_interface(connection=None) -> OpenhabInterface:
    global OH_INTERFACE
    if connection is None:
        return OH_INTERFACE

    OH_INTERFACE = OpenhabInterface(connection)
    return OH_INTERFACE



from HABApp.core.items import Item
from .. import get_openhab_interface


class RollershutterItem(Item):

    def set_state(self, new_state) -> bool:
        if new_state == 'UP':
            new_state = 0.0
        if new_state == 'DOWN':
            new_state = 100.0
        assert isinstance(new_state, (int, float)) or new_state is None, new_state
        return super().set_state(new_state)

    def up(self):
        """Move shutter up"""
        get_openhab_interface().send_command(self.name, 'UP')

    def down(self):
        """Move shutter down"""
        get_openhab_interface().send_command(self.name, 'DOWN')

    def percent(self, percent: float):
        """Command shutter to value (in percent)

        :param percent: target position in percent
        :return:
        """
        assert 0 <= percent <= 100
        get_openhab_interface().send_command(self.name, str(percent))

    def __str__(self):
        return self.state



import datetime

from HABApp.core.items import Item, ColorItem
from . import SwitchItem, ContactItem, RollershutterItem, DimmerItem


def map_items(name, openhab_type : str, openhab_value : str):
    assert isinstance(openhab_type, str), type(openhab_type)
    assert isinstance(openhab_value, str), type(openhab_value)

    value = openhab_value
    if openhab_value == 'NULL' or openhab_value == 'UNDEF':
        value = None

    # Specific classes
    if openhab_type == "Switch":
        return SwitchItem(name, value)

    if openhab_type == "Contact":
        return ContactItem(name, value)

    if openhab_type == "Rollershutter":
        if value is None:
            return RollershutterItem(name, value)
        return RollershutterItem(name, float(value))

    if openhab_type == "Dimmer":
        return DimmerItem(name, value)

    if openhab_type == "Number":
        if value is None:
            return Item(name, value)

        # Number items can be int or float
        try:
            return Item(name, int(value))
        except ValueError:
            return Item(name, float(value))

    if openhab_type == "DateTime":
        if value is None:
            return Item(name, value)
        dt = datetime.datetime.strptime(value.replace('+', '000+'), '%Y-%m-%dT%H:%M:%S.%f%z')
        # all datetimes from openhab have a timezone set so we can't easily compare them
        # TypeError: can't compare offset-naive and offset-aware datetimes
        dt = dt.astimezone(tz=None)   # Changes datetime object so it uses system timezone
        dt = dt.replace(tzinfo=None)  # Removes timezone awareness
        return Item(name, dt)

    if openhab_type == "Color":
        if value is None:
            return ColorItem(name)
        return ColorItem(name, *(float(k) for k in value.split(',')))

    return Item(name, value)



ITEM_TYPES = [
    'String', 'Number', 'Switch', 'Contact', 'Dimmer', 'Rollershutter',
    'Color', 'Contact', 'DateTime', 'Location', 'Player', 'Group']

GROUP_FUNCTIONS = ['AND', 'OR', 'NAND', 'NOR', 'AVG', 'MAX', 'MIN', 'SUM']


import itertools
import logging
import typing

from HABApp.util import PrintException
from . import EventBusListener

_event_log = logging.getLogger('HABApp.EventBus')
_habapp_log = logging.getLogger('HABApp')


_EVENT_LISTENER: typing.Dict[str, typing.List[EventBusListener]] = {}
_EVENT_LISTENER_ALL_EVENTS: typing.List[EventBusListener] = []


def __get_listener_description(listener: EventBusListener) -> str:
    if listener.name is None:
        return f'all names (type {listener.event_filter})'
    else:
        return f'"{listener.name}" (type {listener.event_filter})'


@PrintException
def post_event(name, event):

    _event_log.info(event)

    # Notify all listeners
    for listener in itertools.chain(_EVENT_LISTENER.get(name, []), _EVENT_LISTENER_ALL_EVENTS):
        listener.notify_listeners(event)

    return None


def add_listener(listener: EventBusListener):
    assert isinstance(listener, EventBusListener)
    add_to_all = listener.name is None

    item_listeners = _EVENT_LISTENER.setdefault(listener.name, []) if not add_to_all else _EVENT_LISTENER_ALL_EVENTS

    # don't add the same listener twice
    if listener in item_listeners:
        _habapp_log.warning(f'Event listener for {__get_listener_description(listener)} has already been added!')
        return None

    # add listener
    item_listeners.append( listener)
    _habapp_log.debug(f'Added event listener for {__get_listener_description(listener)}')
    return None


def remove_listener(listener: EventBusListener):
    assert isinstance(listener, EventBusListener)
    add_to_all = listener.name is None

    item_listeners = _EVENT_LISTENER.get(listener.name, []) if not add_to_all else _EVENT_LISTENER_ALL_EVENTS

    # print warning if we try to remove it twice
    if listener not in item_listeners:
        _habapp_log.warning(f'Event listener for {__get_listener_description(listener)} has already been removed!')
        return None

    # remove listener
    item_listeners.remove(listener)
    _habapp_log.debug(f'Removed event listener for {__get_listener_description(listener)}')


def remove_all_listeners():
    _EVENT_LISTENER.clear()
    _EVENT_LISTENER_ALL_EVENTS.clear()



import typing

from HABApp.core.items.item import Item as __Item


_ALL_ITEMS: typing.Dict[str, __Item] = {}


class ItemNotFoundException(Exception):
    pass


def item_exists(name: str) -> bool:
    return name in _ALL_ITEMS


def get_item(name: str) -> __Item:
    try:
        return _ALL_ITEMS[name]
    except KeyError:
        raise ItemNotFoundException(f'Item {name} does not exist!')


def get_all_items() -> typing.List[__Item]:
    return list(_ALL_ITEMS.values())


def get_item_names() -> typing.List[str]:
    return list(_ALL_ITEMS.keys())


def create_item(name: str, item_factory, item_state=None) -> __Item:
    assert issubclass(item_factory, __Item), item_factory
    _ALL_ITEMS[name] = new_item = item_factory(name, state=item_state)
    return new_item


def set_item(item: __Item):
    assert isinstance(item, __Item), type(item)
    _ALL_ITEMS[item.name] = item


def pop_item(name: str) -> __Item:
    return _ALL_ITEMS.pop(name)



from .wrapper import PrintException

from .timeframe import TimeFrame

from .counter import Counter
from .period_counter import PeriodCounter
from .threshold import Threshold
from .statistics import Statistics
from .prioritized_value import PrioritizedValue


import asyncio
import datetime
import logging
import random
import sys
import traceback
import typing
# import warnings

import HABApp
import HABApp.core
import HABApp.openhab
import HABApp.rule_manager
import HABApp.util
from HABApp.core.events import AllEvents
from .interfaces import async_subprocess_exec
from .rule_parameter import RuleParameter
from .scheduler import ReoccurringScheduledCallback, ScheduledCallback, WorkdayScheduledCallback, \
    WeekendScheduledCallback, DayOfWeekScheduledCallback, TYPING_DATE_TIME, TYPING_TIME
from .watched_item import WatchedItem

log = logging.getLogger('HABApp.Rule')


class Rule:
    def __init__(self):

        # get the variables from the caller
        depth = 1
        while True:
            __vars = sys._getframe(depth).f_globals
            depth += 1
            if '__HABAPP__RUNTIME__' in __vars:
                __runtime__ = __vars['__HABAPP__RUNTIME__']
                __rule_file__ = __vars['__HABAPP__RULE_FILE__']
                break

        # variable vor unittests
        test = __vars.get('__UNITTEST__', False)

        # this is a list which contains all rules of this file
        __vars['__HABAPP__RULES'].append(self)

        assert isinstance(__runtime__, HABApp.runtime.Runtime)
        self.__runtime: HABApp.runtime.Runtime = __runtime__

        if not test:
            assert isinstance(__rule_file__, HABApp.rule_manager.RuleFile)
        self.__rule_file: HABApp.rule_manager.RuleFile = __rule_file__

        self.__event_listener: typing.List[HABApp.core.EventBusListener] = []
        self.__future_events: typing.List[ScheduledCallback] = []
        self.__watched_items: typing.List[WatchedItem] = []
        self.__unload_functions: typing.List[typing.Callable[[], None]] = []
        # schedule cleanup
        self.register_on_unload(self.__cleanup_rule)


        # suggest a rule name if it is not
        self.rule_name: str = self.__rule_file.suggest_rule_name(self)

        # interfaces
        self.async_http: HABApp.rule.interfaces.AsyncHttpConnection = self.__runtime.async_http if not test else None
        self.mqtt = self.__runtime.mqtt_connection.interface if not test else None
        self.oh: HABApp.openhab.OpenhabInterface = HABApp.openhab.get_openhab_interface() if not test else None
        self.openhab: HABApp.openhab.OpenhabInterface = self.oh

    def __cleanup_rule(self):
        # Important: set the dicts to None so we don't schedule a future event during _cleanup.
        # If dict is set to None we will crash instead but it is no problem because everything gets unloaded anyhow
        event_listeners = self.__event_listener
        future_events = self.__future_events

        self.__event_listener = None
        self.__future_events = None
        self.__watched_items = None

        # Actually remove the listeners/events
        for listener in event_listeners:
            HABApp.core.EventBus.remove_listener(listener)

        for event in future_events:
            event.cancel()
        return None

    def item_exists(self, name: str) -> bool:
        """
        Checks whether an item exists

        :param name: item name
        :return: True or False
        """
        assert isinstance(name, str), type(name)
        return HABApp.core.Items.item_exists(name)

    def get_item(self, name: str, item_factory=None) -> HABApp.core.items.Item:
        """
        Return the item with the specified name or create the item if it doesn't exist.

        :param name: name of the item
        :param item_factory: if specified and the item does not exist an item
                             with the item_factory-class will be created
        :return: item
        """
        if item_factory is None:
            return HABApp.core.Items.get_item(name)

        try:
            return HABApp.core.Items.get_item(name)
        except HABApp.core.Items.ItemNotFoundException:
            return HABApp.core.Items.create_item(name, item_factory)

    def get_item_state(self, name: str, default=None) -> typing.Any:
        """
        Return the state of the item.

        :param name: item name
        :param default: If the item does not exist or is None this value will be returned (has to be != None)
        :return: state of the specified item
        """
        if default is None:
            return HABApp.core.Items.get_item(name).state

        try:
            state = HABApp.core.Items.get_item(name).state
        except HABApp.core.Items.ItemNotFoundException:
            return default

        if state is None:
            return default
        return state

    def set_item_state(self, name: str, value: typing.Any):
        """
        Set a new state for an item. State can be anything so it is possible to set custom class instances
        and load them in another rule. Using this will also generate the appropriate event.

        If the item doesn't exist it will be created

        :param name: item name or item instance
        :param value: value for new item state
        """
        if isinstance(name, str):
            try:
                item = HABApp.core.Items.get_item(name)
            except HABApp.core.Items.ItemNotFoundException:
                item = HABApp.core.Items.create_item(name, HABApp.core.items.Item)
        else:
            assert isinstance(name, HABApp.core.items.Item)
            item = name

        item.post_state(value)
        return None

    def item_watch(self, name: typing.Union[str, HABApp.core.items.Item],
                   seconds_constant: int, watch_only_changes=True) -> WatchedItem:
        """
        | Keep watch on the state of an item.
        | if `watch_only_changes` is True (default) and the state does not change for `seconds_constant` a
          `ValueNoChangeEvent` will be sent to the event bus.
        | if `watch_only_changes` is False and the state does not receive and update for `seconds_constant` a
          `ValueNoUpdateEvent` will be sent to the event bus.

        :param name: item name or item that shall be watched
        :param seconds_constant: the amount of seconds the item has to be constant or has not received an update
        :param watch_only_changes:
        """
        assert isinstance(name, (str, HABApp.core.items.Item)), type(name)
        assert isinstance(seconds_constant, int)
        assert isinstance(watch_only_changes, bool)

        item = WatchedItem(
            name=name.name if isinstance(name, HABApp.core.items.Item) else name,
            constant_time=seconds_constant,
            watch_only_changes=watch_only_changes
        )
        self.__watched_items.append(item)
        return item

    def item_watch_and_listen(self, name: str, seconds_constant: int, callback,
                              watch_only_changes=True) -> typing.Tuple[WatchedItem, HABApp.core.EventBusListener]:
        """
        Convenience function which combines :class:`~HABApp.Rule.item_watch` and :class:`~HABApp.Rule.listen_event`

        :param name: item name
        :param seconds_constant:
        :param callback: callback that accepts one parameter which will contain the event
        :param watch_only_changes:
        :return:
        """

        watched_item = self.item_watch(name, seconds_constant, watch_only_changes)
        event_listener = self.listen_event(
            name,
            callback,
            HABApp.core.events.ValueNoChangeEvent if watch_only_changes else HABApp.core.events.ValueNoUpdateEvent
        )
        return watched_item, event_listener

    def post_event(self, name, event):
        """
        Post an event to the event bus

        :param name: name or item to post event to
        :param event: Event class to be used (must be class instance)
        :return:
        """
        assert isinstance(name, (str, HABApp.core.items.Item)), type(name)
        return HABApp.core.EventBus.post_event(name.name if isinstance(name, HABApp.core.items.Item) else name, event)

    def listen_event(self, name: typing.Union[HABApp.core.items.Item, str, None], callback,
                     even_type: typing.Union[AllEvents, typing.Any] = AllEvents
                     ) -> HABApp.core.EventBusListener:
        """
        Register an event listener

        :param name: item or name to listen to. Use None to listen to all events
        :param callback: callback that accepts one parameter which will contain the event
        :param even_type: Event filter. This is typically :class:`~HABApp.core.ValueUpdateEvent` or
            :class:`~HABApp.core.ValueChangeEvent` which will also trigger on changes/update from openhab
            or mqtt.
        """
        cb = HABApp.core.WrappedFunction(callback, name=self.__get_rule_name(callback))
        listener = HABApp.core.EventBusListener(
            name.name if isinstance(name, HABApp.core.items.Item) else name, cb, even_type
        )
        self.__event_listener.append(listener)
        HABApp.core.EventBus.add_listener(listener)
        return listener

    def execute_subprocess(self, callback, program, *args, capture_output=True):
        """Run another program

        :param callback: |param_scheduled_cb| after process has finished. First parameter will
                         be an instance of :class:`~HABApp.rule.FinishedProcessInfo`
        :param program: program or path to program to run
        :param args: |param_scheduled_cb_args|
        :param capture_output: Capture program output, set to `False` to only capture return code
        :return:
        """

        assert isinstance(program, str), type(program)
        cb = HABApp.core.WrappedFunction(callback, name=self.__get_rule_name(callback))

        asyncio.run_coroutine_threadsafe(
            async_subprocess_exec(cb.run, program, *args, capture_output=capture_output),
            self.__runtime.loop  # this has to be passed because we will not call it from the main thread
        )

    def run_every(self, time: TYPING_TIME, interval, callback, *args, **kwargs) -> ReoccurringScheduledCallback:
        """
        Run a function periodically

        :param time: |param_scheduled_time|
        :param interval:
        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        cb = HABApp.core.WrappedFunction(callback, name=self.__get_rule_name(callback))
        future_event = ReoccurringScheduledCallback(time, interval, cb, *args, **kwargs)
        self.__future_events.append(future_event)
        return future_event

    def run_on_day_of_week(self, time: TYPING_TIME, weekdays, callback, *args, **kwargs) -> DayOfWeekScheduledCallback:
        """

        :param time: |param_scheduled_time|
        :param weekdays:
        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """

        # names of weekdays in local language
        lookup = {datetime.date(2001, 1, i).strftime('%A'): i for i in range(1, 8)}
        lookup.update( {datetime.date(2001, 1, i).strftime('%A')[:3]: i for i in range(1, 8)})

        # abbreviations in German and English
        lookup.update({"Mo": 1, "Di": 2, "Mi": 3, "Do": 4, "Fr": 5, "Sa": 6, "So": 7})
        lookup.update({"Mon": 1, "Tue": 2, "Wed": 3, "Thu": 4, "Fri": 5, "Sat": 6, "Sun": 7})
        lookup = {k.lower(): v for k, v in lookup.items()}

        if isinstance(weekdays, int) or isinstance(weekdays, str):
            weekdays = [weekdays]
        for i, val in enumerate(weekdays):
            if not isinstance(val, str):
                continue
            weekdays[i] = lookup[val.lower()]

        cb = HABApp.core.WrappedFunction(callback, name=self.__get_rule_name(callback))
        future_event = DayOfWeekScheduledCallback(time, weekdays, cb, *args, **kwargs)
        self.__future_events.append(future_event)
        return future_event

    def run_on_every_day(self, time: TYPING_TIME, callback, *args, **kwargs) -> DayOfWeekScheduledCallback:
        """

        :param time: |param_scheduled_time|
        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        cb = HABApp.core.WrappedFunction(callback, name=self.__get_rule_name(callback))
        future_event = DayOfWeekScheduledCallback(time, [1, 2, 3, 4, 5, 6, 7], cb, *args, **kwargs)
        self.__future_events.append(future_event)
        return future_event

    def run_on_workdays(self, time: TYPING_TIME, callback, *args, **kwargs) -> WorkdayScheduledCallback:
        """

        :param time: |param_scheduled_time|
        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        cb = HABApp.core.WrappedFunction(callback, name=self.__get_rule_name(callback))
        future_event = WorkdayScheduledCallback(time, cb, *args, **kwargs)
        self.__future_events.append(future_event)
        return future_event

    def run_on_weekends(self, time: TYPING_TIME, callback, *args, **kwargs) -> WeekendScheduledCallback:
        """

        :param time: |param_scheduled_time|
        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        cb = HABApp.core.WrappedFunction(callback, name=self.__get_rule_name(callback))
        future_event = WeekendScheduledCallback(time, cb, *args, **kwargs)
        self.__future_events.append(future_event)
        return future_event

    def run_daily(self, callback, *args, **kwargs) -> ReoccurringScheduledCallback:
        """
        Picks a random hour, minute and second and runs the callback every day

        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        start = datetime.timedelta(seconds=random.randint(0, 24 * 3600 - 1))
        interval = datetime.timedelta(days=1)
        return self.run_every(start, interval, callback, *args, **kwargs)

    def run_hourly(self, callback, *args, **kwargs) -> ReoccurringScheduledCallback:
        """
        Picks a random minute and second and run the callback every hour

        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        start = datetime.timedelta(seconds=random.randint(0, 3600 - 1))
        interval = datetime.timedelta(seconds=3600)
        return self.run_every(start, interval, callback, *args, **kwargs)

    def run_minutely(self, callback, *args, **kwargs) -> ReoccurringScheduledCallback:
        """
        Picks a random second and runs the callback every minute

        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        start = datetime.timedelta(seconds=random.randint(0, 60 - 1))
        interval = datetime.timedelta(seconds=60)
        return self.run_every(start, interval, callback, *args, **kwargs)

    def run_at(self, date_time: TYPING_DATE_TIME, callback, *args, **kwargs) -> ScheduledCallback:
        """
        Run a function at a specified date_time

        :param date_time:
        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        cb = HABApp.core.WrappedFunction(callback, name=self.__get_rule_name(callback))
        future_event = ScheduledCallback(date_time, cb, *args, **kwargs)
        self.__future_events.append(future_event)
        return future_event

    def run_in(self, seconds: int, callback, *args, **kwargs) -> ScheduledCallback:
        """
        Run the callback in x seconds

        :param int seconds: Wait time in seconds before calling the function
        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        assert isinstance(seconds, int), f'{seconds} ({type(seconds)})'

        cb = HABApp.core.WrappedFunction(callback, name=self.__get_rule_name(callback))
        future_event = ScheduledCallback(datetime.timedelta(seconds=seconds), cb, *args, **kwargs)
        self.__future_events.append(future_event)
        return future_event

    def run_soon(self, callback, *args, **kwargs) -> ScheduledCallback:
        """
        Run the callback as soon as possible (typically in the next second).

        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        cb = HABApp.core.WrappedFunction(callback, name=self.__get_rule_name(callback))
        future_event = ScheduledCallback( None, cb, *args, **kwargs)
        self.__future_events.append(future_event)
        return future_event

    def get_rule_parameter(self, file_name: str, *keys, default_value='ToDo') -> RuleParameter:
        """

        :param file_name: Name of the file (without extension), will get created if it doesn't exist
        :param keys: section name which value shall be loaded
        :param default_value: if the corresponding entry in the file does not exist
                              it will be created with default_value
        """
        assert isinstance(file_name, str), type(file_name)
        return RuleParameter(self.__runtime.rule_params, file_name, *keys, default_value=default_value)

    def get_rule(self, rule_name: str) -> 'typing.Union[Rule, typing.List[Rule]]':
        assert rule_name is None or isinstance(rule_name, str), type(rule_name)
        return self.__runtime.rule_manager.get_rule(rule_name)

    def register_on_unload(self, func):
        """Register a function with no parameters which will be called when the rule is unloaded.
        Use this for custom cleanup functions.

        :param func: function which will be called
        """
        assert callable(func)
        assert func not in self.__unload_functions, 'Function was already registered!'
        self.__unload_functions.append(func)

    # -----------------------------------------------------------------------------------------------------------------
    # internal functions
    # -----------------------------------------------------------------------------------------------------------------
    def __get_rule_name(self, callback):
        return f'{self.rule_name}.{callback.__name__}' if self.rule_name else None

    @HABApp.util.PrintException
    def _check_rule(self):

        # Check if items do exists
        if not HABApp.core.Items.get_all_items():
            return None

        for listener in self.__event_listener:
            # Listener listens to all changes
            if listener.name is None:
                continue

            # check if specific item exists
            if not HABApp.core.Items.item_exists(listener.name):
                log.warning(f'Item "{listener.name}" does not exist (yet)! '
                            f'self.listen_event in "{self.rule_name}" may not work as intended.')

        for item in self.__watched_items:
            if not HABApp.core.Items.item_exists(item.name):
                log.warning(f'Item "{item.name}" does not exist (yet)! '
                            f'self.item_watch in "{self.rule_name}" may not work as intended.')

    @HABApp.util.PrintException
    def _process_events(self, now):

        # watch items
        clean_items = False
        for item in self.__watched_items:
            item.check(now)
            if item.is_canceled:
                clean_items = True
        if clean_items:
            self.__watched_items = [k for k in self.__watched_items if not k.is_canceled]

        # sheduled events
        clean_events = False
        for future_event in self.__future_events:  # type: ScheduledCallback
            future_event.check_due(now)
            future_event.execute()
            if future_event.is_finished:
                clean_events = True

        # remove finished events
        if clean_events:
            self.__future_events = [k for k in self.__future_events if not k.is_finished]
        return None

    @HABApp.util.PrintException
    def _unload(self):

        # unload all functions
        for func in self.__unload_functions:
            try:
                func()
            except Exception as e:

                # try getting function name
                try:
                    name = f' in "{func.__name__}"'
                except AttributeError:
                    name = ''

                log.error( f'Error{name} while unloading "{self.rule_name}": {e}')

                # log traceback
                lines = traceback.format_exc().splitlines()
                del lines[1:3]  # see implementation in wrappedfunction.py why we do this
                for line in lines:
                    log.error(line)




import HABApp.core

from .base_event import BaseItemEvent


class ChannelTriggeredEvent(BaseItemEvent, HABApp.core.events.ValueUpdateEvent):
    def __init__(self, _in_dict):
        super().__init__(_in_dict)

        # smarthome/items/NAME/state
        self.name: str = self._topic[19:-10]
        self.event: str = self._payload['event']
        self.channel: str = self._payload['channel']

        # value wird gesetzt, weil wir von ValueUpdateEvent ableiten.
        # Todo: überlegen ob es nicht eine EventKlasse gibt
        self.value = self.event

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, event: {self.event}>'



import datetime
import typing
import HABApp


class Item:

    @classmethod
    def get_create_item(cls, name: str, default_state=None):
        """Creates a new item in HABApp and returns it or returns the already existing one with the given name

        :param name: item name
        :param default_state: state the item will have if it gets created
        :return: item
        """
        assert isinstance(name, str), type(name)

        try:
            item = HABApp.core.Items.get_item(name)
        except HABApp.core.Items.ItemNotFoundException:
            item = cls(name, default_state)
            HABApp.core.Items.set_item(item)

        assert isinstance(item, cls), f'{cls} != {type(item)}'
        return item

    def __init__(self, name: str, state=None):
        assert isinstance(name, str), type(name)

        self.name: str = name
        self.state: typing.Any = state

        _now = datetime.datetime.now()
        self.last_change: datetime.datetime = _now
        self.last_update: datetime.datetime = _now

    def set_state(self, new_state) -> bool:
        """Set a new state without creating events

        :param new_state: new state
        :return: True if state has changed
        """
        state_changed = self.state != new_state

        _now = datetime.datetime.now()
        if state_changed:
            self.last_change = _now
        self.last_update = _now

        self.state = new_state
        return state_changed

    def post_state(self, new_state):
        """Set a new state and post appropriate events on the event bus (``ValueUpdateEvent``, ``ValueChangeEvent``)

        :param new_state: new state
        """
        old_state = self.state
        self.set_state(new_state)

        # create events
        HABApp.core.EventBus.post_event(self.name, HABApp.core.events.ValueUpdateEvent(self.name, new_state))
        if old_state != new_state:
            HABApp.core.EventBus.post_event(
                self.name, HABApp.core.events.ValueChangeEvent(self.name, value=new_state, old_value=old_state)
            )
        return None

    def get_state(self, default_value=None) -> typing.Any:
        """Return the state of the item.

        :param default_value: Return this value if the item state is None
        :return: State of the item
        """
        if self.state is None:
            return default_value
        return self.state

    def __repr__(self):
        ret = ''
        for k in ['name', 'state', 'last_change', 'last_update']:
            ret += f'{", " if ret else ""}{k}: {getattr(self, k)}'
        return f'<{self.__class__.__name__} {ret:s}>'

    # only support == and != operators by default
    # __ne__ delegates to __eq__ and inverts the result so this is not overloaded separately
    def __eq__(self, other):
        return self.state == other

    def __bool__(self):
        return bool(self.state)

    # rich comparisons only for numeric types (int and float)
    def __lt__(self, other):
        if not isinstance(self.state, (int, float)):
            return NotImplemented
        return self.state < other

    def __le__(self, other):
        if not isinstance(self.state, (int, float)):
            return NotImplemented
        return self.state <= other

    def __ge__(self, other):
        if not isinstance(self.state, (int, float)):
            return NotImplemented
        return self.state >= other

    def __gt__(self, other):
        if not isinstance(self.state, (int, float)):
            return NotImplemented
        return self.state > other



import threading


class Counter:
    """Implements a thread safe counter"""

    def __init__(self, initial_value: int = 0, on_change = None):
        """

        :param initial_value: Initial value of the counter
        :param on_change:  Function which will be called when the counter changes.
                           First argument will be the new counter value.
                           Function will also be called when the Counter gets created.
        """
        assert isinstance(initial_value, int)

        self.__lock = threading.Lock()

        self.__initial_value = initial_value
        self.__value = self.__initial_value

        # func which gets called when the counter changes
        self.on_change = on_change
        if self.on_change:
            self.on_change(self.__value)

    @property
    def value(self) -> int:
        """Current value"""
        return self.__value

    def reset(self):
        """Reset value to initial value"""
        with self.__lock:
            changed = self.__value != self.__initial_value
            self.__value = self.__initial_value
            value = self.__value

        if changed and self.on_change:
            self.on_change(value)

    def increase(self, step=1) -> int:
        """Increase value

        :param step: increase by this value, default = 1
        :return: value of the counter
        """
        with self.__lock:
            self.__value += step
            value = self.__value

        if self.on_change:
            self.on_change(value)
        return value

    def decrease(self, step=1) -> int:
        """Decrease value

        :param step: decrease by this value, default = 1
        :return: value of the counter
        """
        with self.__lock:
            self.__value -= step
            value = self.__value

        if self.on_change:
            self.on_change(value)

        return value
